/* -*- mode: c++; c-basic-offset: 3; -*- */
#include "ldas/ldasconfig.hh"

#include "chan_config.hh"
#include "frame_source.hh"
#include "fzstream.hh"
#include "framecpp/Common/IOStream.hh"
#include "framecpp/IFrameStream.hh"
#include "framecpp/FrTOC.hh"
#include "framecpp/FrameH.hh"
#include "framecpp/FrAdcData.hh"
#include "framecpp/FrRawData.hh"
#include "framecpp/FrProcData.hh"
#include "framecpp/FrVect.hh"
#if LDAS_VERSION_NUMBER >= 200000
#include "ldastoolsal/fstream.hh"
#else
#include "general/fstream.hh"
#endif
#include <fstream>
#include <sstream>
#include <iostream>
#include <memory>

//======================================  Define pointer types
#if LDAS_VERSION_NUMBER > 119010
#define FCPP_SHARED_PTRS 1
#if LDAS_VERSION_NUMBER >= 200000
#include "ldastoolsal/SharedPtr.hh"
typedef LDASTools::AL::SharedPtr<FrameCPP::FrameH>     frame_pointer;
typedef LDASTools::AL::SharedPtr<FrameCPP::FrRawData>  frrawdata_pointer;
typedef LDASTools::AL::SharedPtr<FrameCPP::FrVect>     frvect_pointer;
#else
#include "general/SharedPtr.hh"
typedef General::SharedPtr<FrameCPP::FrameH>     frame_pointer;
typedef General::SharedPtr<FrameCPP::FrRawData>  frrawdata_pointer;
typedef General::SharedPtr<FrameCPP::FrVect>     frvect_pointer;
#endif
#else
typedef FrameCPP::FrameH*     frame_pointer;
typedef FrameCPP::FrRawData*  frrawdata_pointer;
typedef FrameCPP::FrVect*     frvect_pointer;
#endif

typedef FrameCPP::FrRawData::const_firstAdc_iterator AdcDataIterator;
typedef FrameCPP::FrameH::const_procData_iterator    ProcDataIterator;
#if LDAS_VERSION_NUMBER >= 200000
typedef LDASTools::AL::filebuf filebuffer_type;
#else
typedef General::filebuf filebuffer_type;
#endif 

using namespace std;
using namespace sends;

//======================================  Chan config constructors
chan_config::chan_config(void) 
   : _start(0), _stop(0), _debug(0), _compress(cls_none)
{
}

chan_config::chan_config(const string& frtype, gps_type start, gps_type stop, 
			 const chan_list& chlist) 
   : _type(frtype), _start(start), _stop(stop), _debug(0), _compress(cls_none),
     _list(chlist)
{}

//======================================  Chan config destructors
chan_config::~chan_config(void) {
}

//======================================  Guess the channel type
chan_type
GuessChanType(const string& name, double rate, chan_type deftype) {
   string::size_type dot = name.rfind(".");
   if (dot != string::npos) {
      string ext = name.substr(dot+1);
      if (ext == "mean" || ext == "rms" || ext == "n" || 
	  ext == "min" || ext == "max") {
	 if (rate < 1.0) return chan_type::kMinuteTrend;
	 else            return chan_type::kSecondTrend;
      }
   }
   return deftype;
}

//======================================  Remove ugly characters from the name
static void
FixName(std::string& name) {
   string::size_type inx = name.find_first_of(" \t\n\r\b");
   while (inx != string::npos) {
      name.erase(inx, 1);
      inx = name.find_first_of(" \t\n\r\b");
   }
}


//======================================  Read a channel list
void
chan_config::getChanList(const std::string& pfx, gps_type gps) {
   chan_entry null_ent(string(), chan_type::kUnknown);

   //----------------------------------  Initialize segment values.
   _list.clear();
   _start = gps;
   _stop  = gps+1;
   _type  = pfx;
   if (_debug) cerr << "Get channel list for type: " << _type << " GPS: "
		    << gps << endl;

   //----------------------------------  Get the frame path from Source_List.
   //                                    If no frames exist, return the gap.
   const frame_loc* fl = Source_List.find(pfx, gps);
   if (!fl) {
      frame_source::gpsseg_type gap;
      Source_List.find_gap(pfx, gps, gap);
      _start = gap.first;
      _stop  = gap.second;
      return;
   }

   //----------------------------------  Open a frame reader. Bail on error.
   string file;
   fl->file_path(gps).swap(file);
   //cout << "Scan: " << _type << " at: " << gps << " file: " << file << endl;
   FrameCPP::Common::FrameBuffer< filebuffer_type >* 
      fbuf = new FrameCPP::Common::FrameBuffer<filebuffer_type>(ios::in);
   if (!fbuf->open(file.c_str(), ios::in)) {
      cerr << "Unable to open file: " << file << endl;
      return;
   } else if (_debug) {
      cerr << "Opened file: " << file << endl;
   }
   FrameCPP::IFrameStream reader(fbuf);

   //----------------------------------  Count the channels, preallocate a
   //                                    temporary channel list.
   const FrameCPP::Common::FrTOC* toc = reader.GetTOC();
   index_type nChans = toc->nameADC().size() + toc->nameProc().size();
   //cout << "Total channels: " << nChans << endl;
   index_type nFr    = toc->nFrame();
   chan_list::chan_vect temp_list;

   //-----------------------------------  Loop over frames
   for (index_type iFr=0; iFr<nFr; iFr++) {
      temp_list.clear();
      temp_list.reserve(nChans);
 
      //--------------------------------  Read the whole frame.
      //unsigned int rflgs = FrameCPP::FrameH::RAW_DATA;
      frame_pointer frame = reader.ReadNextFrame();

      //--------------------------------  Get frame time interval
      //                                    Fix known frame bug - GTime == 0!
      gps_type start = frame->GetGTime().GetSeconds();
      if (!start) start = fl->frame_start(gps);
      double dt = frame->GetDt();
      gps_type stop  = start + gps_type(dt);

      //--------------------------------  Get raw (FrAdcData) data
      data_type dtype;
      frrawdata_pointer raw = frame->GetRawData();
      if (raw) {
	 for (AdcDataIterator itr = raw->RefFirstAdc().begin();
	      itr != raw->RefFirstAdc().end(); ++itr) {

	    //--------------------------  Get channel parameters.
	    string name((*itr)->GetName());
	    FixName(name);
	    double rate = (*itr)->GetSampleRate();
	    FrameCPP::FrVect& v = *((*itr)->RefData().front());
	    rate = 1.0/v.GetDim(0).GetDx();
	    chan_type ctype = GuessChanType(name, rate, chan_type::kRawData);
	    dtype.set_frame_type(v.GetType());

	    //--------------------------  Add channel to the temporary list
	    chan_entry ce(name, ctype, 0, dtype, rate, _type);
	    temp_list.push_back(null_ent);
	    temp_list.back().swap(ce);
	 }
      }

      //--------------------------------  Get processed data
      for (ProcDataIterator itr = frame->RefProcData().begin();
	   itr != frame->RefProcData().end(); ++itr) {

	 //-----------------------------  Get channel parameters.
	 string name((*itr)->GetName());
	 FixName(name);
	 frvect_pointer v = (*itr)->RefData().front();
	 double rate = 1.0/v->GetDim(0).GetDx();
	 //cout << "Channel: " << name << " rate: " << rate << endl;
	 chan_type ctype = GuessChanType(name, rate, chan_type::kReduced);
	 dtype.set_frame_type(v->GetType());

	 //-----------------------------  Add channel to the temporary list
	 chan_entry ce(name, ctype, 0, dtype, rate, _type);
	 temp_list.push_back(null_ent);
	 temp_list.back().swap(ce);
      }
      //--------------------------------  Set time range.
#ifndef FCPP_SHARED_PTRS
      delete frame;
#endif

      //--------------------------------  Merge temporary into channel list.
      _list.merge(temp_list);
      if (start < _start) _start = start;
      if (stop > _stop)   _stop  = stop;
   } // end for(iFr=0 - nFr)
   if (nFr > 1) _list.compress();
}

//======================================  Compare channel lists
bool 
chan_config::operator==(const chan_config& node) const {
   index_type N = _list.size();
   if (node._list.size() != N) return false;
   for (index_type i=0; i<N; ++i) {
      if (node._list[i] != _list[i]) return false;
   }
   return true;
}

//======================================  Merge time segments.
void 
chan_config::merge(const chan_config& node) {
   if (node._start < _start) _start = node._start;
   if (_stop  <  node._stop) _stop  = node._stop;
}

//======================================  Merge time segments.
std::string
chan_config::range(void) const {
   ostringstream ostr;
   ostr << start() << "-" << stop();
   return ostr.str();
}

//======================================  Set the compression mode
void
chan_config::set_compress(cls_compress mode) {
   _compress = mode;
}

//======================================  Set the debug level
void
chan_config::set_debug(unsigned long debug) {
   _debug = debug;
}

//======================================  Write the channel list.
void
chan_config::write(const std::string& cdir) const {

   //----------------------------------  Generrate a file name
   ostringstream ostr;
   if (!cdir.empty()) {
      ostr << cdir;
      if (cdir[cdir.size()-1] != '/') ostr << "/";
   }
   ostr << _type << "-" << _start << "-" << _stop - _start << ".cls";

   //----------------------------------  Open an outpu stream
   std::unique_ptr<std::ostream> file;
   switch (_compress) {
   case cls_none:
      file.reset(new ofstream(ostr.str().c_str()));
      break;
   case cls_gzip:
      ostr << ".gz";
      file.reset(new fz_ostream(ostr.str().c_str()));
      break;
   }

   //----------------------------------  Write the list
   index_type N = _list.size();
   for (index_type i=0; i<N; i++) {
      _list[i].write(*file);
   }
}
